#include "mpi.h"
#include <algorithm>
#include <chrono>
#include <iostream>
#include <stdlib.h>

#define MASTER 0

using namespace std;

// Function to perform LU factorization in parallel using MPI
// U - Upper triangular matrix (modified in-place)
// L - Lower triangular matrix
// P - Permutation matrix
// size - Size of the square matrix
// numtask - Number of tasks (MPI processes)
void LU_factorization(float **U, float **L, float **P, int size, int numtask) {
    MPI_Request temp_request;

    for (int k = 0; k < size; ++k) {
        int size_sent_matrix = size - k;   // Size of submatrix to be sent
        int number_lines = size - (k + 1); // Number of rows below the current pivot row

        // Check for singularity of the matrix
        if (U[k][k] == 0) {
            throw runtime_error("Matrix is singular!");
        }

        // Compute factors for L matrix
        for (int i = k + 1; i < size; i++) {
            double factor = U[i][k] / U[k][k];
            L[i][k] = factor;
        }

        // Distribute computation of updates to rows among worker processes
        for (int t = 0; t < numtask - 1; ++t) {
            int start = t * (double)number_lines / (numtask - 1);
            int a = (int)((t + 1) * (double)number_lines / (numtask - 1));
            int end = (a > number_lines) ? number_lines : a;

            int total_lines = end - start + 1;

            // Send relevant data to workers
            MPI_Isend(&total_lines, 1, MPI_INT, t + 1, 0, MPI_COMM_WORLD, &temp_request);
            MPI_Isend(&(U[k][k]), 1, MPI_FLOAT, t + 1, 1, MPI_COMM_WORLD, &temp_request);
            MPI_Isend(&size_sent_matrix, 1, MPI_INT, t + 1, 2, MPI_COMM_WORLD, &temp_request);
            MPI_Isend(&(U[k][k]), size_sent_matrix, MPI_FLOAT, t + 1, 100, MPI_COMM_WORLD,
                      &temp_request);

            for (int i = start; i < end; ++i) {
                MPI_Isend(&(U[k + i + 1][k]), size_sent_matrix, MPI_FLOAT, t + 1,
                          100 + i - start + 1, MPI_COMM_WORLD, &temp_request);
            }
        }

        // Receive updated rows from worker processes
        for (int t = 0; t < numtask - 1; ++t) {
            int start = t * (double)number_lines / (numtask - 1);
            int a = (int)((t + 1) * (double)number_lines / (numtask - 1));
            int end = (a > number_lines) ? number_lines : a;
            int total_lines = end - start;
            for (int i = start; i < end; ++i) {
                MPI_Recv(&(U[k + i + 1][k]), size_sent_matrix, MPI_FLOAT, t + 1,
                         100 + i - start + 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            }
        }
    }

    // Notify workers to exit
    int exit_comm = -1;
    for (int t = 1; t < numtask; ++t) {
        MPI_Send(&exit_comm, 1, MPI_INT, t, 0, MPI_COMM_WORLD);
    }
}

// Worker function for parallel computation
void work(int matrix_size, int rank) {
    float **matrix = (float **)malloc(matrix_size * sizeof(*matrix));
    if (!matrix) {
        fprintf(stderr, "Malloc failed.\n");
        exit(1);
    }

    for (int i = 0; i < matrix_size; ++i) {
        matrix[i] = (float *)malloc(matrix_size * sizeof(**matrix));
        if (!matrix[i]) {
            fprintf(stderr, "Malloc failed.\n");
            exit(1);
        }
    }

    int command;
    while (1) {
        MPI_Recv(&command, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        if (command == -1) {
            break; // Exit when receiving the termination signal
        }

        int lines = command;
        float u_k_k;
        MPI_Recv(&u_k_k, 1, MPI_FLOAT, 0, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        int size;
        MPI_Recv(&size, 1, MPI_INT, 0, 2, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

        // Receive the matrix data from the master process
        for (int i = 0; i < lines; ++i) {
            MPI_Recv(&(matrix[i][0]), size, MPI_FLOAT, 0, i + 100, MPI_COMM_WORLD,
                     MPI_STATUS_IGNORE);
        }

        // Perform row updates locally
        for (int i = 1; i < lines; ++i) {
            float factor = matrix[i][0] / u_k_k;
            for (int j = 0; j < size; ++j) {
                matrix[i][j] -= factor * matrix[0][j];
            }
        }

        // Send updated rows back to the master process
        for (int i = 1; i < lines; ++i) {
            MPI_Send(matrix[i], size, MPI_FLOAT, 0, i + 100, MPI_COMM_WORLD);
        }
    }

    // Free allocated memory
    for (int i = 0; i < matrix_size; ++i) {
        free(matrix[i]);
    }
    free(matrix);
}

int main(int argc, char *argv[]) {
    int numtasks, rank, len;
    char hostname[MPI_MAX_PROCESSOR_NAME];

    // Initialize MPI environment
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Get_processor_name(hostname, &len);

    int matrix_size;
    if (rank == MASTER) {
        if (argc != 3) {
            fprintf(stderr, "Please provide the correct number of arguments.");
            exit(-1);
        }

        string out_directory = "../out/";
        FILE *f = fopen(argv[1], "r");

        int size_matrix;
        fscanf(f, "%d", &size_matrix);
        matrix_size = size_matrix;

        MPI_Bcast(&matrix_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
        fflush(stdout);

        // Allocate matrices
        float **P, **U, **L, **A;
        A = (float **)malloc(size_matrix * sizeof(*A));
        L = (float **)malloc(size_matrix * sizeof(*L));
        U = (float **)malloc(size_matrix * sizeof(*U));
        P = (float **)malloc(size_matrix * sizeof(*P));

        if (!A || !L || !U || !P) {
            fprintf(stderr, "Malloc failed");
            exit(-1);
        }

        for (int i = 0; i < size_matrix; ++i) {
            A[i] = (float *)malloc(size_matrix * sizeof(**A));
            L[i] = (float *)malloc(size_matrix * sizeof(**L));
            U[i] = (float *)malloc(size_matrix * sizeof(**U));
            P[i] = (float *)malloc(size_matrix * sizeof(**P));
        }

        for (int i = 0; i < size_matrix; ++i) {
            for (int j = 0; j < size_matrix; ++j) {
                int aux;
                fscanf(f, "%d", &(aux));
                A[i][j] = (float)aux;
                U[i][j] = (float)aux;
            }
            L[i][i] = 1.0;
            P[i][i] = 1.0;
        }

        fclose(f);

        // Start timer
        auto start = std::chrono::high_resolution_clock::now();

        // Perform LU factorization
        LU_factorization(U, L, P, size_matrix, numtasks);

        // Stop timer
        auto stop = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
        std::cerr << "[INFO] LU MPI: " << duration.count() << "us\n";

        // Write results to the output file
        out_directory.append("mpi_");
        out_directory.append(argv[2]);
        f = fopen(out_directory.c_str(), "w");

        // Write lower triangular matrix L to the output file
        fprintf(f, "L:\n");
        for (int i = 0; i < size_matrix; ++i) {
            for (int j = 0; j < size_matrix; ++j) {
                fprintf(f, "%f ", L[i][j]);
            }
            fprintf(f, "\n");
        }
        fprintf(f, "\n");

        // Write upper triangular matrix U to the output file
        fprintf(f, "U:\n");
        for (int i = 0; i < size_matrix; ++i) {
            for (int j = 0; j < size_matrix; ++j) {
                fprintf(f, "%f ", U[i][j]);
            }
            fprintf(f, "\n");
        }

        // Free allocated memory
        for (int i = 0; i < size_matrix; ++i) {
            free(A[i]);
            free(L[i]);
            free(U[i]);
            free(P[i]);
        }
        free(A);
        free(L);
        free(U);
        free(P);
        fclose(f);
    } else {
        // Workers handle parallel computation
        MPI_Bcast(&matrix_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
        work(matrix_size, rank);
    }

    // Finalize MPI environment
    MPI_Finalize();
    return 0;
}