#include <algorithm>
#include <chrono>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <thread>
#include <vector>

// Worker function for LU decomposition
// Processes a subset of rows (from 'start' to 'end') for the current step 'k'
// U - Upper triangular matrix (modified in-place)
// L - Lower triangular matrix
// N - Size of the matrix
// k - Current step
void workers(float **U, float **L, int N, int k, int start, int end) {
    // i > k always, i is divided to threads
    // i> k => U[k][k] constant
    // i > k => U[i][k] singular access, row access no cache
    // i > k => L[i][k] singular access, row acces no cache
    // i > k && j >= k => U[k][j] is already completed constant
    // i > k && j >= k => U[i][j] is distributed to the current thread only.

    for (int i = start; i < end; ++i) {
        if (U[k][k] == 0)
            throw std::runtime_error("Matrix is singular!");

        // Compute the multiplier factor for row elimination
        float factor = U[i][k] / U[k][k];
        L[i][k] = factor;

        // Update the remaining elements in the row
        for (int j = k; j < N; ++j)
            U[i][j] -= factor * U[k][j];
    }
}

// Function to perform LU factorization using pthreads
// U - Upper triangular matrix (modified in-place)
// L - Lower triangular matrix
// N - Size of the matrix
// num_threads - Number of threads to use
void LU_pthreads(float **U, float **L, int N, int num_threads) {
    if (num_threads > N)
        throw std::runtime_error("Oversubscribing");

    for (int k = 0; k < N; ++k) {
        std::vector<std::thread> threads;

        // Divide the rows among threads
        for (int t_idx = 0; t_idx < num_threads; ++t_idx) {
            int start = (k + 1) + t_idx * (double)(N - (k + 1)) / num_threads;
            int end =
                std::min((k + 1) + (t_idx + 1) * (double)(N - (k + 1)) / num_threads, (double)N);

            threads.emplace_back(workers, U, L, N, k, start, end);
        }

        // Wait for all threads to finish
        for (auto &thread : threads)
            thread.join();
    }
}

int main(int argc, char **argv) {
    // Validate command-line arguments
    if (argc != 4) {
        std::cerr << "Usage: ./lu_pthreads <input_file> <output_suffix> <num_threads>\n";
        return EXIT_FAILURE;
    }

    std::string out_directory = "../out/";

    // Open input file
    std::ifstream fin(argv[1]);
    if (!fin) {
        std::cerr << "Error opening file " << argv[1] << '\n';
        return EXIT_FAILURE;
    }

    int num_threads = std::stoi(argv[3]);

    int N;
    fin >> N;

    // Allocate memory for matrices L and U
    float **L = new float *[N];
    float **U = new float *[N];

    for (int i = 0; i < N; ++i) {
        L[i] = new float[N];
        U[i] = new float[N];
    }

    float val;
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < N; ++j) {
            fin >> val;
            U[i][j] = val;
            L[i][j] = 0.0;
        }

        L[i][i] = 1.0;
    }

    fin.close();

    // Start timer
    auto start = std::chrono::high_resolution_clock::now();
    // Perform LU factorization
    LU_pthreads(U, L, N, num_threads);
    // Stop timer
    auto stop = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
    std::cerr << "[INFO] LU pthreads: " << duration.count() << "us\n";

    // Prepare output file
    out_directory.append("pthreads_");
    out_directory.append(argv[2]);
    std::ofstream fout(out_directory);
    if (!fout) {
        std::cerr << "Error opening output file" << std::endl;
        return EXIT_FAILURE;
    }

    fout << std::fixed << std::setprecision(6);

    fout << "L:\n";
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < N; ++j) {
            fout << L[i][j] << " ";
        }
        fout << "\n";
    }
    fout << "\n";

    fout << "U:\n";
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < N; ++j) {
            fout << U[i][j] << " ";
        }
        fout << "\n";
    }
    fout << "\n";

    fout.close();

    // Free allocated memory
    for (int i = 0; i < N; ++i) {
        delete[] L[i];
        delete[] U[i];
    }

    delete[] L;
    delete[] U;

    return EXIT_SUCCESS;
}